'''key components needed for RFLoss'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import torch
from torch.nn.modules.loss import _Loss
from torch import Tensor

from .layers import RMSNormalizeIQ


class RFLoss(_Loss):
    '''
    Loss criterion for complex RF timeseries.
    '''
    def __init__(self,
            spatial_size:int=4096,
            weight_spec:float=1, weight_xcor:float=1,
            weight_align:float=1, weight_env:float=1, weight_fft:float=0,
            size_average=None, reduce=None, reduction:str='mean', data_format='ncl') -> None:
        '''
        This criterion was the result of a long experimentation phase and
        tries to balance the loss from each component of the calculation.

        Any weights that equal zero will skip computation.

        Parameters
        ----------
        spatial_size : int, default 4096
            The length of the input complex tensors. If omitted the weights between metrics will no longer be uniform.
        weight_spec : float, default 1
            Valuable metric that compares inputs with multiple spectrograms at various resolutions.
        weight_xcor : float, default 1
            Valuable metric that estimates real & imaginary time-domain error.
        weight_align: float, default 1
            Metric that emphasizes the importance of the DC alignment, a component of the full xcor.
        weight_env : float, default 1
            OK metric that ensure the scale of inputs w.r.t phase are similar over the observation period.
            In my experience ML autoencoders learn this quickly then do not improve.
        weight_fft : float, default 0
            This component disabled since it is simply a much worse version of the spectrogram loss.
        data_format : str, default 'ncl'
            Network normally consumes and produces complex-valued data represented as real-valued (NCL)
            but if data is complex-valued (NL) will add a transform layer during encode/decode.

        Similarity Metrics
        ------------------
        * Are the signals in phase? -> env_loss
        * Are the signals similar in magnitude? -> env_loss
        * Are the signals similar in time domain? -> xcor_loss
        * Are the signals aligned in time domain? -> align_loss
        * Are the signals similar in freq domain? -> spec_loss or fft_loss
        * Are the signals similar in 2D spectral representation? -> spec_loss

        Error between Complex Waveforms
        -------------------------------
        In order to implement mean complex error we can use:
            1) mean((abs(a)-abs(b))**2) # compare magnitude only
            2) mean(abs(a-b)**2) == mean((a.real-b.real)**2 + (a.imag-b.imag)**2) # compare phase & mag
            3) mean(a*a.conj() - b*b.conj()) # worse magnitude comparison

        Notes
        -----
        * Beware Magnitude & Phase comparisons made within env_loss are un-normalized
        * Spectrograms computed at multiple resolutions on normalized (apple, banana).
          Where `blah = log2(len(ray_apple)`,
          I think appropriate fftsizes are `[2**(blah-2), 2**(blah-1), 2**blah, 2**(blah+1), 2**(blah+2)]`.
          Keep both magnitude and complex portion of error; complex portion hard to figure in early epochs.
          Ensures frequency domain correlation will work at multiple scales.
        * Normalized Complex Correlation loss ensures time domain crosscorrelation will be correct
        '''
        super().__init__(size_average, reduce, reduction)

        spatial_exponent = torch.log2(torch.tensor(spatial_size))
        assert spatial_exponent % 1 == 0, 'RFLoss can only be quickly computed on 2**n length tensors'
        assert data_format in ['ncl', 'nl']
        self.fftsizes = torch.nn.Parameter(2**(torch.arange(-2, 3, dtype=int) + int(spatial_exponent / 2)), requires_grad=False)
        self.spec_scale = torch.nn.Parameter(torch.tensor(2 / len(self.fftsizes)), requires_grad=False)
        self.spatial_size = torch.nn.Parameter(torch.tensor(2**spatial_exponent), requires_grad=False)
        self.weight_xcor = torch.nn.Parameter(torch.tensor(weight_xcor), requires_grad=False)
        self.weight_align = torch.nn.Parameter(torch.tensor(weight_align), requires_grad=False)
        self.weight_spec = torch.nn.Parameter(torch.tensor(weight_spec), requires_grad=False)
        self.weight_env = torch.nn.Parameter(torch.tensor(weight_env), requires_grad=False)
        self.weight_fft = torch.nn.Parameter(torch.tensor(weight_fft), requires_grad=False)
        self._rms = RMSNormalizeIQ(spatial_size=spatial_size)
        self.data_format = data_format

    def __repr__(self):
        return '{}(fftsizes={}, weights=({:.2f}, {:.2f}, {:.2f}, {:.2f}, {:.2f}), {})'.format(
            type(self).__name__,
            '['+' '.join('{}'.format(ffz) for ffz in self.fftsizes)+']',
            self.weight_xcor.item(),
            self.weight_align.item(),
            self.weight_spec.item(),
            self.weight_env.item(),
            self.weight_fft.item(),
            self.data_format.upper(),
        )

    @staticmethod
    def spec_turbo(ray, fftsize=None):
        '''
        *fastest* low-resolution spectrogram for use in loss function
        no overlap for this version!

        Parameters
        ----------
        ray : torch.Tensor
            ray can be (batchsize, complex_1d_tensor) or (complex_1_tensor) shape.
        fftsize : int, optional
            Determine # of columns in spectrogram. Should specified for speed, but if
            not will automatically determine a square-ish shape.

        Returns
        -------
        spec : torch.Tensor
            2d tensor with un-fftshifted spectrogram.

        Benchmark
        ---------
        00.9μs for batch size 64
        '''
        ray_len = ray.shape[-1]
        if fftsize is None:
            # using round() instead of int() would yield more square, but we want smaller spec
            fftsize = 2**(int(torch.log2(torch.tensor(ray_len))) // 2)
        block_len = torch.div(ray_len, fftsize, rounding_mode='floor')
        # doing reshape this way loses a few samples if ray_len / fftsize is not an integer!
        spec = ray[..., 0:fftsize*block_len].reshape((-1, block_len, fftsize))
        spec = torch.fft.fft(spec, n=fftsize)
        # note that spec is unshifted and complex still
        return spec

    def calc_spec_loss(self, apple_cplx:Tensor, banana_cplx:Tensor):
        '''
        loss from various spectrogram sizes

        Returns
        -------
        loss : Tensor
            Value in range(0, 1).
            Will be 1 if completely uncorrelated.
            Will be 0 if completely correlated.
        '''
        spec_loss = 0
        for fftsize in self.fftsizes:
            spec_apple = self.spec_turbo(apple_cplx, fftsize=fftsize)
            spec_banana = self.spec_turbo(banana_cplx, fftsize=fftsize)
            # comparison does not include phase
            spec_loss += torch.mean((torch.abs(spec_apple)-torch.abs(spec_banana))**2) / fftsize
        return spec_loss * self.spec_scale

    def calc_xcor_loss(self, apple_fft:Tensor, banana_fft:Tensor):
        '''
        cross-correlation loss

        track correlation of both real & imag to reconstruct phase

        Returns
        -------
        xcor_loss : Tensor
            Good measure of rough time alignment.
            Value in range(0, 1).
            Will be 1 if completely uncorrelated.
            Will be 0 if completely correlated at any time offset within slice.
        align_loss : Tensor
            Good measure of fine time alignment.
            Value in range(0, 1).
            Will be 1 if perfectly aligned in time.
            Will be 0 if no relationship.
        '''
        # Peak value of xcor will be close to 1 if same pulse in both slices
        corr = torch.abs(torch.fft.ifft(apple_fft * torch.conj(banana_fft))) / self.spatial_size
        xcor_loss = 1 - torch.mean(torch.max(corr, dim=-1).values)
        # DC component of correlation should be maximized if aligned
        align_loss = 1 - torch.mean(torch.abs(corr[..., 0]))
        return xcor_loss, align_loss

    def calc_env_loss(self, apple_cplx:Tensor, banana_cplx:Tensor):
        '''
        Envelope Loss
        Computed before normalization!

        Returns
        -------
        loss : Tensor
            Loss in range(0, 2) if similar scale; in range(0, inf) if unbounded.
            Will be 4 if oppositely correlated
            Will be 2 if completely uncorrelated.
            Will be <1 if similar scale, likely floating around 0.5.
            Will be 0 if identical.
        '''
        # div2 to be similar to MSELoss
        return torch.mean(torch.abs(apple_cplx - banana_cplx)**2) * .5

    def calc_fft_loss(self, apple_fft:Tensor, banana_fft:Tensor):
        '''
        compute loss from fullsize fft

        Returns
        -------
        loss : Tensor
            Value in range(0, 1).
            Will be 1 if freq domain uncorrelated.
            Will be 0 if freq domain looks identical.
        '''
        return torch.mean((torch.abs(apple_fft) - torch.abs(banana_fft))**2) / (2 * self.spatial_size)

    def forward(self, apple:Tensor, banana:Tensor) -> Tensor:
        '''
        compute loss and metrics

        If initialized with data_format == ncl, assume complex data in ray is
        real-valued (batchsize, 2, len) to represent complex-valued data channels first
        benchmark: 4096 len is 14 ms using batch_size=64 (223μs per row)

        Parameters
        ----------
        apple : torch.Tensor
            Tensor with shape (batchsize, 2, spatial_size) that will be converted to torch.complex64.
        banana : torch.Tensor
            See apple.
        '''
        if self.data_format == 'ncl':
            apple_cplx = torch.view_as_complex(apple.swapaxes(-1, -2).contiguous())
            banana_cplx = torch.view_as_complex(banana.swapaxes(-1, -2).contiguous())
        else:
            apple_cplx = apple
            banana_cplx = banana

        metrics = {}
        total_loss = torch.tensor(0, device=apple.device, dtype=torch.float)
        if self.weight_env:
            env_loss = self.weight_env * self.calc_env_loss(apple_cplx, banana_cplx)
            metrics['env_loss'] = env_loss
            total_loss += env_loss
        # normalize for other loss considerations
        apple_cplx = self._rms(apple_cplx)
        banana_cplx = self._rms(banana_cplx)
        if self.weight_spec:
            spec_loss = self.weight_spec * self.calc_spec_loss(apple_cplx, banana_cplx)
            metrics['spec_loss'] = spec_loss
            total_loss += spec_loss
        if self.weight_xcor or self.weight_align or self.weight_fft:
            # observe freq domain if necessary
            apple_fft = torch.fft.fft(apple_cplx)
            banana_fft = torch.fft.fft(banana_cplx)
        if self.weight_xcor or self.weight_align:
            xcor_loss, align_loss = self.calc_xcor_loss(apple_fft, banana_fft)
            # weight * .25 to make similar magnitude to other components
            if self.weight_xcor:
                xcor_loss *= self.weight_xcor * .25
                metrics['xcor_loss'] = xcor_loss
                total_loss += xcor_loss
            if self.weight_align:
                align_loss *= self.weight_align * .25
                metrics['align_loss'] = align_loss
                total_loss += align_loss
        if self.weight_fft:
            fft_loss = self.weight_fft * self.calc_fft_loss(apple_fft, banana_fft)
            metrics['fft_loss'] = fft_loss
            total_loss += fft_loss

        metrics['loss'] = total_loss
        return total_loss, metrics



# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import logging
import math
from collections import namedtuple
import numpy as np

import torch
import lightning as pl
from torch import nn

from .layers import DropConnect

log = logging.getLogger(__name__)

# BlockArgs are used to specify a Glaucus architecture, which is thankfully generated smartly by `blockgen`.
BlockArgs = namedtuple('BlockArgs', [
    'num_repeat', 'kernel_size', 'stride', 'filters_in', 'filters_out', 'expand_ratio', 'squeeze_ratio'])


def blockgen(
        spatial_in:int=4096, spatial_out:int=8,
        filters_in:int=2, filters_out:int=64,
        expand_ratio:int=4, squeeze_ratio:int=4,
        steps:int=6, mode:str='encoder'):
    '''
    smartly generate block sequence

    * Geometrically increase/decrease filters and increase repeats.
    * Distribute spatial scaling across units
    * Geometrically increase kernel size from 1.5% to 15% per [1][2][3][4].

    After a week long experiment in Oct 2021 I determined that you always want
    strides to be larger earlier in either encoder or decoder network.
    Similarly for number of repeat steps you want more repeats as the network gets deeper.

    References
    ----------
    [1] https://arxiv.org/abs/2104.00298
    [2] https://arxiv.org/abs/1905.11946
    [3] https://arxiv.org/abs/2109.07161
    [4] https://arxiv.org/abs/1905.02244
    '''
    # functions that round UP to either even or odd
    round_to_odd = lambda val: np.ceil(val) // 2 * 2 + 1
    round_to_even = lambda val: np.ceil(val / 2) * 2
    # calculate required strides; smartly distribute scaling
    strides = np.ones(steps, dtype=int)
    doubles = np.abs(np.log2(spatial_in/spatial_out))
    assert doubles % 1 == 0, 'spatial values must be 2**n'
    for dbx in range(int(doubles)):
        # smartly distribute strides to achieve spatial size change
        strides[dbx % steps] *= 2
    # calculate the fraction of spatial time-series covered by kernel
    # between geomspace and linspace per [1][2][3][4]
    kernel_percent = (np.linspace(.015, .15, steps) + np.geomspace(.015, .15, steps))/2
    if mode == 'decoder':
        kernel_percent = np.flip(kernel_percent)
    # calculate the filters per step
    filter_steps = np.geomspace(filters_in, filters_out, steps+1).round().astype(int)

    # calculate the repeats per block; 15 max, for now use 8
    repeat_steps = np.geomspace(2, 8, steps).round().astype(int)
    blocks = []
    params = np.zeros(steps) # only used for logging
    for sdx in range(steps):
        if mode == 'encoder':
            spatial_in /= strides[sdx]
        else:
            spatial_in *= strides[sdx]
        kernel_size = int(max(3, round_to_odd(kernel_percent[sdx]*spatial_in)))
        while kernel_size < strides[sdx]:
            # this is bad
            # https://ezyang.github.io/convolution-visualizer/index.html
            # https://distill.pub/2016/deconv-checkerboard/
            log.warning('bumping kernel_size to prevent checkerboarding resulting from kernel_size < stride')
            kernel_size += 2

        block = BlockArgs(
            num_repeat=repeat_steps[sdx],
            kernel_size=kernel_size,
            stride=strides[sdx],
            filters_in=filter_steps[sdx],
            filters_out=filter_steps[sdx+1],
            expand_ratio=expand_ratio,
            squeeze_ratio=squeeze_ratio,
        )
        log.info(f'{block}')
        params[sdx] += block.filters_in * block.filters_out # conv_reshape
        params[sdx] += block.filters_in * block.filters_in * block.expand_ratio * block.kernel_size # conv_expand
        params[sdx] += (block.filters_in * block.expand_ratio)**2 # conv_middle
        params[sdx] += (block.filters_in * block.expand_ratio)**2 / block.squeeze_ratio # linear_squeeze x2
        params[sdx] += block.filters_in * block.expand_ratio * block.filters_out # conv_tail
        log.debug(f'params={params[sdx]:.0f}, out_shape=({block.filters_out:.0f}, {spatial_in:.0f})')
        blocks += [block]
    return blocks


# defaults
ENCODER_BLOCKS = blockgen(steps=6, spatial_in=4096, spatial_out=8, filters_in=2, filters_out=64, mode='encoder')
DECODER_BLOCKS = blockgen(steps=6, spatial_in=8, spatial_out=4096, filters_in=64, filters_out=2, mode='decoder')


class GBlock(pl.LightningModule):
    '''
    Inspired by ResNetBlock, MBConvBlock, & FusedMBConvBlock

    Parameters
    ----------
    filters_in : int
        Channel dim of NCL input.
    filters_out : int
        Channel dim of NCL output.
    mode : str, default 'encoder'
        Either 'encoder' or 'decoder', modifies where scaling occurs and
        determines whether 'stride' will decimate or interpolate.
    stride : int, default 1
        When stride is > 1, will decimate or interpolate the output length depending on 'mode'.
    expand_ratio : int, default 4
        Expansion ratio for input channel count.
    squeeze_ratio : float, default 4
        Squeeze excitation ratio applied to expanded channel count.
    kernel_size : int, default 7
        Kernel size should be adjusted based on network depth.
        Note even kernel sizes not supported due to padding difficulties.
    drop_connect_rate : float, default 0.2
        Proability of dropped connections in residual layer.
        With a deep network design this value should be zero early and scale up to max value.

    Key Features
    ------------
    * Scalable Width & Depth [6][7]
    * Fused Inverted Residual Blocks [10][6][7]
    * Stochastic Depth [5]
    * Squeeze & Excitation [8] on expanded channels
    * DropConnect instead of Dropout [9]

    References
    ----------
    [1] https://github.com/lukemelas/EfficientNet-PyTorch/blob/master/efficientnet_pytorch/model.py
    [2] https://github.com/lukemelas/EfficientNet-PyTorch/blob/master/efficientnet_pytorch/utils.py
    [3] https://www.fatalerrors.org/a/efficientnet-v2-papers-and-code-analysis.html
    [4] https://github.com/pytorch/vision/blob/main/torchvision/models/resnet.py
    [5] https://arxiv.org/abs/1603.09382
    [6] https://arxiv.org/abs/2104.00298
    [7] https://arxiv.org/abs/1905.11946
    [8] https://arxiv.org/abs/1709.01507
    [9] http://yann.lecun.com/exdb/publis/pdf/wan-icml-13.pdf
    [10] https://arxiv.org/abs/1905.02244

    Design
    ------
    filters_in ➡ filters_ex ➡ filters_sq ➡ filters_out
    only apply stride on first repeat of each block

    original_stack = conv ➡ bn ➡ activ ➡ conv ➡ bn ➡ +x ➡ activ
    preactiv_stack = bn ➡ activ ➡ conv ➡ bn ➡ activ ➡ conv ➡ +x
    fused_stack = conv_ex ➡ bn ➡ activ ➡ +se ➡ conv ➡ bn ➡ +x

    In EfNetV2, depth is scaled from (1, 5) and width is scaled (1, 4)

    Major Changes
    -------------
    * GBlock design will always add residual unlike efnet.
    * Will include Expand-Squeeze from EfNet.
    * Will include Fused blocks from EFNetV2.
    * Squeeze size calculated on expanded filters, not input filters

    *DANGER* even kernel sizes not quite supported; padding nightmares
    '''
    def __init__(self,
                 filters_in:int, filters_out:int, mode:str='encoder',
                 stride:int=1, drop_connect_rate:float=0.2,
                 expand_ratio:int=4, squeeze_ratio:int=4, kernel_size:int=7):
        super().__init__()
        assert mode in ['encoder', 'decoder']
        self.filters_in = filters_in
        self.filters_out = filters_out
        self.expand_ratio = expand_ratio
        self.squeeze_ratio = squeeze_ratio
        self.mode = mode
        self.stride = stride
        self.bn_mom = 1e-2 # better than torch default
        self.bn_eps = 1e-3 # better than torch default
        self.filters_ex = self.filters_in * self.expand_ratio
        self.filters_sq = max(1, round(self.filters_ex / self.squeeze_ratio))
        self.is_reshaped = self.filters_in != self.filters_out or self.stride != 1
        self.kernel_size = kernel_size
        if kernel_size % 2 == 1:
            # odd kernel size
            self.pad_size = (kernel_size - 1) // 2
        else:
            # this is not quite working
            self.pad_size = (kernel_size // 2) - 1
        # special arguments for convolution steps
        expand_args = {}
        reshape_args = {}
        # self._zpad = torch.nn.ZeroPad2d(padding=(1,0)) # using for 1D when even kernel
        # Define all layers
        self._activ = nn.SiLU(inplace=True) # Activation De-Jour

        # adjust for encoder/decoder
        if self.mode == 'encoder':
            # more depth, less samples
            Conv1d = nn.Conv1d
        else:
            # less depth, more samples
            Conv1d = nn.ConvTranspose1d

        expand_args['kernel_size'] = self.kernel_size
        expand_args['stride'] = self.stride
        expand_args['padding'] = self.pad_size
        if self.mode == 'decoder':
            # tricky deconvolution padding required
            expand_args['output_padding'] = self.stride - 1

        if self.is_reshaped:
            # we need this to adapt the identity to the correct shape
            # using as few weights as possible. Large experiment on 2021-12-27
            # determined k=1 was better than attempting to use k=stride
            if self.mode == 'decoder' and self.stride != 1:
                # tricky deconvolution padding
                reshape_args['output_padding'] = self.stride - 1
            self._reshape = nn.Sequential(
                Conv1d(self.filters_in, self.filters_out, stride=self.stride, bias=False, kernel_size=1, **reshape_args),
                nn.BatchNorm1d(self.filters_out),
            )

        # Expansion Layers
        self._conv_expand = Conv1d(self.filters_in, self.filters_ex, bias=False, **expand_args)
        self._bn0 = nn.BatchNorm1d(self.filters_ex, momentum=self.bn_mom, eps=self.bn_eps)

        if self.squeeze_ratio != 1:
            # Squeeze-Excitation Layers
            self._avgpool = nn.AdaptiveAvgPool1d(1)
            self._se_reduce = nn.Linear(self.filters_ex, self.filters_sq)
            self._se_expand = nn.Linear(self.filters_sq, self.filters_ex)

        self._conv_tail = Conv1d(self.filters_ex, self.filters_out, kernel_size=1, bias=False)
        self._bn1 = nn.BatchNorm1d(self.filters_out, momentum=self.bn_mom, eps=self.bn_eps)
        self._dropconnect = DropConnect(drop_connect_rate)

        # Block Info
        log.debug('GBlock{}(stride={}, filters=({},{},{},{}))'.format(
                'Enc' if self.mode == 'encoder' else 'Dec', self.stride,
                self.filters_in, self.filters_ex, self.filters_sq, self.filters_out,
                ))

    def forward(self, x):
        identity = x
        if self.is_reshaped:
            # we might have different output shape
            identity = self._reshape(identity)
        # Fused Expansion Phase (Inverted Bottleneck)
        x = self._conv_expand(x)
        x = self._bn0(x)
        x = self._activ(x)
        # Squeeze-Excitation Phase
        if self.squeeze_ratio != 1:
            x_squeezed = self._avgpool(x).squeeze()
            x_squeezed = self._se_reduce(x_squeezed)
            x_squeezed = self._activ(x_squeezed)
            x_squeezed = self._se_expand(x_squeezed).unsqueeze(-1)
            # doing this in-place causes backprop err
            x = x * torch.sigmoid(x_squeezed)
        # Pointwise Convolution Phase
        x = self._conv_tail(x)
        x = self._bn1(x)
        # drop connection on some decreasing fraction of x only while training
        x = self._dropconnect(x)
        # Skip Connection
        x += identity
        return x


class GlaucusNet(nn.Module):
    def __init__(self,
                 blocks=ENCODER_BLOCKS, mode:str='encoder',
                 width_coef:float=1.0, depth_coef:float=1.0,
                 spatial_dim:int=4096, drop_connect_rate:float=0.2,
        ) -> None:
        super().__init__()
        assert isinstance(blocks, list), 'should be list of BlockArgs'
        assert len(blocks) > 0, 'cannot have zero blocks'
        assert mode in ['encoder', 'decoder']
        self.width_coef = width_coef
        self.depth_coef = depth_coef
        self.mode = mode
        # drop connect rate is scaled from 0 to drop_connect_rate through block network
        self.drop_connect_rate = drop_connect_rate
        if self.mode == 'decoder':
            # need to reshape flat input into (filters_in, spatial_dim) for blocks
            self.filters_in = blocks[0].filters_in
        # construct blocks
        self._blocks = nn.ModuleList()
        drop_num, drop_denom = 0, -1 # keep track to scale drop_connect_rate
        for bdx, block_args in enumerate(blocks):
            # count up how many total blocks there actually are with change to dept_coef
            drop_denom += self.round_repeats(block_args.num_repeat, self.depth_coef)
        for bdx, block_args in enumerate(blocks):
            # account for width & depth scaling
            num_repeat = self.round_repeats(block_args.num_repeat, self.depth_coef)
            if bdx == 0:
                # fixed stem size on first block input
                blk_filters_in = block_args.filters_in
            else:
                blk_filters_in = self.round_filters(block_args.filters_in, self.width_coef)
            if bdx == len(blocks)-1:
                # fixed stem size on last block output
                blk_filters_out = block_args.filters_out
            else:
                blk_filters_out = self.round_filters(block_args.filters_out, self.width_coef)
            self._blocks.append(GBlock(
                filters_in=blk_filters_in,
                filters_out=blk_filters_out,
                expand_ratio=block_args.expand_ratio,
                squeeze_ratio=block_args.squeeze_ratio,
                kernel_size=block_args.kernel_size,
                stride=block_args.stride, # only first block handles stride
                drop_connect_rate=self.drop_connect_rate*(drop_num/drop_denom),
                mode=self.mode,
            ))
            drop_num += 1
            # handle repeats
            for _ in range(num_repeat - 1):
                self._blocks.append(GBlock(
                    filters_in=blk_filters_out,
                    filters_out=blk_filters_out,
                    expand_ratio=block_args.expand_ratio,
                    squeeze_ratio=block_args.squeeze_ratio,
                    kernel_size=block_args.kernel_size,
                    stride=1, # no scale change in repeats
                    drop_connect_rate=self.drop_connect_rate*(drop_num/drop_denom),
                    mode=self.mode,
                ))
                drop_num += 1

    @staticmethod
    def round_filters(base, multiplier):
        '''vary width based on multiplier'''
        return int(round(base * multiplier))

    @staticmethod
    def round_repeats(base, multiplier):
        '''vary depth based on multiplier'''
        return int(math.ceil(base * multiplier))

    def extract_features(self, x):
        for block in self._blocks:
            x = block(x)
        return x

    def forward(self, x):
        if self.mode == 'decoder':
            # un-flatten to feed into decoder
            x = x.view(x.size(0), self.filters_in, -1)
        x = self.extract_features(x)
        if self.mode == 'encoder':
            # pooling & final linear layer
            # x = self._avgpool(x).squeeze(-1)
            # TODO: Note to self, if reimplementing avgpool, use self.dropout_rate
            x = x.flatten(start_dim=1)
        return x



'''custom layers for pytorch'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import typing
import torch
import lightning as pl


def wrap_ncl(func):
    '''
    decorator that allows the use of NCL (batchsize, 2, spatial_size) RF data
    with modules designed for complex RF NL (batchsize, spatial_size) RF data.
    '''
    def wrap(x_ncl, *args, **kwargs):
        x_nl = torch.view_as_complex(x_ncl.swapaxes(-1, -2).contiguous())
        y_nl = func(x_nl, *args, **kwargs)
        y_ncl = torch.view_as_real(y_nl).swapaxes(-1, -2)
        return y_ncl
    return wrap


class RMSNormalizeIQ(pl.LightningModule):
    '''
    When consuming RF, ensure the waveform has uniform scale to regularize input to our architecture.
    Expects Complex-Valued NL format (batchsize, spatial_size).

    Best Method
    -----------
    With numpy the [vdot]_ method is fastest, but some testing showed that using [norm]_ was best with pytorch.

    | Method   | CPU Time (ms) | GPU Time (µs) |
    |----------|---------------|---------------|
    |  [norm]_ |      10.1     |      203      |
    |  [vdot]_ |      13.3     |      386      |
    | [naive]_ |      14.3     |      462      |

    Tests were run on Threadripper 3990WX and single RTX2080. Input shape was (512, 8192).

    RMS Methods
    -----------
    All these methods find RMS amplitude equally.
    .. [norm] x.norm(p=2, dim=-1, keepdim=True)
    .. [vdot] torch.sqrt(torch.bmm(x.view(batch_size, 1, spatial_size), torch.conj(x.view(batch_size, spatial_size, 1))).squeeze(-1).real)
    .. [naive] torch.sqrt(torch.sum(torch.square(torch.abs(x)), axis=-1, keepdim=True))
    '''
    def __init__(self, spatial_size: int = 4096, bias: bool = False) -> None:
        super().__init__()
        self.bias = bias
        # recall that if something is a parameter it will be learned, and registered as a parameter automatically.
        # dis-similarly if something is a tensor it will not be learned, and needs to be a registered as a buffer to appear in model params
        # self.register_buffer('scale', 1 / torch.sqrt(torch.tensor(spatial_size, dtype=torch.float)))
        # self.scale = torch.nn.Parameter(1 / torch.sqrt(torch.tensor(spatial_size, dtype=torch.float)), requires_grad=False)
        self.scale = 1 / torch.sqrt(torch.tensor(spatial_size, dtype=torch.float))
        if bias:
            self.offset = torch.nn.Parameter(torch.zeros(spatial_size, dtype=torch.float))

    def forward(self, x):
        '''
        Returns the original complex signal scaled appropriately.
        '''
        rms = x.norm(p=2, dim=-1, keepdim=True) * self.scale
        x_rms = x / rms
        if self.bias:
            x_rms += self.offset
        return x_rms


class RMSNormalize(RMSNormalizeIQ):
    '''Inherits RMSNormalize, but expects real-valued NCL format (batchsize, 2, spatial_size)'''
    def __init__(self, spatial_size: int = 4096, bias: bool = False):
        super().__init__(spatial_size, bias)
        self.forward = wrap_ncl(self.forward)


class TimeDomain2FreqDomain(pl.LightningModule):
    '''
    Convert Time Domain RF to Freq Domain RF.
    Expects NCL format (batchsize, 2, spatial_size).
    '''
    def __init__(self, fft_shift: bool = True, norm: typing.Optional[str] = None):
        super().__init__()
        if norm and norm not in ['forward', 'backward', 'ortho']:
            raise ValueError(f"Invalid normalization option: {norm}")
        self.fft_shift = fft_shift
        self.norm = norm

    def forward(self, x):
        '''
        fftshift is not strictly required, but keeps features spatially close for potential convolutional layers.
        '''
        x = torch.view_as_complex(x.swapaxes(-1, -2).contiguous())
        x = torch.fft.fft(x, dim=-1, norm=self.norm)
        if self.fft_shift:
            x = torch.fft.fftshift(x, dim=-1)
        x = torch.view_as_real(x).swapaxes(-1, -2)
        return x


class FreqDomain2TimeDomain(pl.LightningModule):
    '''
    Convert Freq Domain RF to Time Domain RF.
    Expects NCL format (batchsize, 2, spatial_size).
    '''
    def __init__(self):
        super().__init__()

    def forward(self, x):
        '''
        fftshift is not strictly required, but keeps features spatially close for potential convolutional layers.
        '''
        x = torch.view_as_complex(x.swapaxes(-1, -2).contiguous())
        x = torch.view_as_real(torch.fft.ifft(torch.fft.fftshift(x, dim=-1))).swapaxes(-1, -2)
        return x


class DropConnect(pl.LightningModule):
    '''
    drop connections between blocks (alternative to dropout)

    The intended use of dropconnect is within residual connections like:
    `out = in + dropconnect(layer(in))`

    Some implementations utilize activation after this step, but we have not.

    Parameters
    ----------
    inputs : torch.Tensor
        Of shape (batchsize, ...).
    drop_connect_rate : float
        Probability of drop connections (0-1).
    training : bool
        Only do this if we are in `training` mode.

    References
    ----------
    [1] http://yann.lecun.com/exdb/publis/pdf/wan-icml-13.pdf
    [2] https://github.com/tensorflow/tpu/blob/cd433314cc6f38c10a23f1d607a35ba422c8f967/models/official/efficientnet/utils.py#L146
    '''
    def __init__(self, drop_connect_rate:float=0.2):
        super().__init__()
        assert 0 <= drop_connect_rate <= 1, 'drop_connect_rate must be in range of [0, 1]'
        self.survival_rate = 1 - drop_connect_rate

    def __repr__(self):
        ''' This makes things look nice when you print the architecture.'''
        return f'{type(self).__qualname__}(survival_rate={self.survival_rate:.2f})'

    def forward(self, x):
        if not self.training:
            # only drop connections during training
            return x
        elif self.survival_rate == 1:
            return x
        else:
            batchsize = x.shape[0]
            # create binary tensor mask
            random_tensor = self.survival_rate + torch.rand((batchsize,) + (1,)*(x.dim()-1), dtype=x.dtype, device=x.device)
            binary_tensor = torch.floor(random_tensor)
            # Unlike canonical method to multiply survival_rate at test time, here we
            # divide survival_rate at training time, such that no addition compute is needed at test time.
            return x / self.survival_rate * binary_tensor


class GaussianNoise(torch.nn.Module):
    '''
    Add gaussian noise to RF.
    Expects NCL format like (batchsize, 2, spatial_size).

    Input should be RMS normalized.
    Returns RMS normalized output.
    '''
    def __init__(self, spatial_size:int=4096, min_snr_db:float=-3, max_snr_db: float = 20):
        super().__init__()
        self.min_snr_db = min_snr_db
        self.max_snr_db = max_snr_db
        self._rms_norm = RMSNormalize(spatial_size=spatial_size)

    def forward(self, x):
        if self.training:
            snr_db = torch.rand(1, device=x.device)[0] * (self.max_snr_db - self.min_snr_db) + self.min_snr_db
            snr = 10**(snr_db/10)
            # unit noise already RMS normalized
            unit_noise = torch.randn(x.size(), device=x.device, dtype=x.dtype) * 0.7071067811865476
            # scale signal and noise
            x = unit_noise + x * snr
            # as much as I want to use something like this, it's not quite the same as normalizing again
            # unit_noise / (snr + 1) + x * snr / (snr + 1)
            x = self._rms_norm(x)

        if self.training:
            x = x, snr_db
        else:
            x = x, None
        return x



'''ensure AEs are working'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import torch

from glaucus import GlaucusAE, FullyConnectedAE


class TestAE(unittest.TestCase):
    def test_ae_roundtrip(self):
        '''the  output size should always be the same as the input size'''
        for AE in [GlaucusAE, FullyConnectedAE]:
            for data_format in ['ncl', 'nl']:
                for domain in ['time', 'freq']:
                    # note if we use a diff spatial_size, will need to gen new encoder & decoder bocks
                    spatial_size = 4096
                    if data_format == 'ncl':
                        trash_x = torch.randn(7, 2, spatial_size)
                    else:
                        trash_x = torch.randn(7, spatial_size, dtype=torch.complex64)
                    ae = AE(domain=domain, data_format=data_format)
                    trash_y, _ = ae(trash_x)
                    self.assertEqual(trash_x.shape, trash_y.shape)

    def test_ae_quantization(self):
        '''If quantization enabled, should use quint8 as latent output'''
        for AE in [FullyConnectedAE, GlaucusAE]:
            for data_format in ['ncl', 'nl']:
                for is_quantized in [True, False]:
                    target = torch.quint8 if is_quantized else torch.float32
                    # note if we use a diff spatial_size, will need to gen new encoder & decoder bocks
                    spatial_size = 4096
                    if data_format == 'ncl':
                        trash_x = torch.randn(7, 2, spatial_size)
                    else:
                        trash_x = torch.randn(7, spatial_size, dtype=torch.complex64)
                    ae = AE(bottleneck_quantize=is_quantized, data_format=data_format)
                    if is_quantized:
                        # this will prepare the quant/dequant layers
                        torch.quantization.prepare(ae, inplace=True)
                        # this applies the quantization coefficients within the bottleneck
                        torch.quantization.convert(ae.cpu(), inplace=True)
                    _, trash_latent = ae(trash_x)
                    self.assertEqual(trash_latent.dtype, target)

    def test_ae_backprop(self):
        '''catch errors during backpropagation'''
        for data_format in ['ncl', 'nl']:
            for AE in [FullyConnectedAE, GlaucusAE]:
                for is_quantized in [True, False]:
                    # note if we use a diff spatial_size, will need to gen new encoder & decoder bocks
                    spatial_size = 4096
                    if data_format == 'ncl':
                        trash_x = torch.randn(7, 2, spatial_size)
                    else:
                        trash_x = torch.randn(7, spatial_size, dtype=torch.complex64)
                    ae = AE(bottleneck_quantize=is_quantized, data_format=data_format)
                    if is_quantized:
                        # this will prepare the quant/dequant layers
                        torch.quantization.prepare(ae, inplace=True)
                        # this applies the quantization coefficients within the bottleneck
                        torch.quantization.convert(ae.cpu(), inplace=True)
                    loss, _ = ae.step((trash_x, None), 0)
                    # will raise RuntimeError here if there is an issue with backprop
                    loss.backward()



# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import logging
import numpy as np

import torch
import lightning as pl

log = logging.getLogger(__name__)


class FullyConnected(pl.LightningModule):
    '''Sequential Layer Generator for 1D Fully Connected'''
    def __init__(self, size_in:int=512, size_out:int=128, steps:int=3,
                 quantize_in:bool=False, quantize_out:bool=False, use_dropout:bool=False) -> None:
        '''
        Smartly construct fully connected of arbitrary depth

        quantize_in means input will be dequantized from uint8
        quantize_out means output is quantized into uint8
        '''
        super().__init__()
        self.save_hyperparameters()
        self.size_in = size_in
        self.size_out = size_out
        self.steps = steps
        self.step_sizes = np.round(np.geomspace(self.size_in, self.size_out, self.steps+1)).astype(int)
        self.quantize_in = quantize_in
        self.quantize_out = quantize_out
        self.bn_mom = 1e-2 # better than torch default
        self.bn_eps = 1e-3 # better than torch default

        # deal with optional quantization
        if self.quantize_in or self.quantize_out:
            # 'fbgemm' is for servers, 'qnnpack' is for mobile
            qconfig = torch.quantization.get_default_qconfig('fbgemm')
        if self.quantize_in:
            self._dequant_in = torch.quantization.DeQuantStub(qconfig=qconfig)
        if self.quantize_out:
            self._quant_out = torch.quantization.QuantStub(qconfig=qconfig)
        if use_dropout:
            self._dropout = torch.nn.Dropout(0.2)

        # define layers
        self._activ = torch.nn.SiLU(inplace=True)
        self._fc = self._make_fc(use_dropout)

        # layer summary
        summary = quantize_in * 'dequant,'
        for sdx in range(steps+1):
            summary += f'{self.step_sizes[sdx]}{"," * (sdx!=steps)}'
        summary += quantize_out * ',quant'
        log.info('FullyConnected(%s)', summary)

    def _make_fc(self, use_dropout=False):
        '''constructor for stepped architecture'''
        layers = []
        if self.quantize_in:
            layers.append(self._dequant_in)
        for sdx in range(self.steps):
            layers.append(torch.nn.Linear(self.step_sizes[sdx], self.step_sizes[sdx+1]))
            layers.append(torch.nn.BatchNorm1d(self.step_sizes[sdx+1], momentum=self.bn_mom, eps=self.bn_eps))
            layers.append(self._activ)
            if use_dropout and sdx != self.steps - 1:
                # do not add dropout on last layer
                layers.append(self._dropout)
        if self.quantize_out:
            layers.append(self._quant_out)
        return torch.nn.Sequential(*layers)

    def forward(self, x):
        x_hat = self._fc(x)
        return x_hat



'''ensure blocks are working'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import torch
from hypothesis import settings, given, strategies as st

from glaucus import GlaucusNet, blockgen, FullyConnected, GBlock


class TestParams(unittest.TestCase):
    '''autoencoders should operate over all valid params'''
    @settings(deadline=None, max_examples=100)
    @given(
        exponent=st.integers(min_value=2, max_value=14),
        steps=st.integers(min_value=1, max_value=8),
        filters_mid=st.integers(min_value=1, max_value=100)
    )
    def test_io_glaucusnet(self, exponent, steps, filters_mid):
        '''design works on a variety of spatial sizes'''
        spatial_dim = 2**exponent
        # for spatial_dim in 2**np.arange(8, 14):
        encoder_blocks = blockgen(steps=steps, spatial_in=spatial_dim, spatial_out=8, filters_in=2, filters_out=filters_mid, mode='encoder')
        decoder_blocks = blockgen(steps=steps, spatial_in=8, spatial_out=spatial_dim, filters_in=filters_mid, filters_out=2, mode='decoder')
        encoder = GlaucusNet(mode='encoder', blocks=encoder_blocks, spatial_dim=spatial_dim)
        decoder = GlaucusNet(mode='decoder', blocks=decoder_blocks, spatial_dim=spatial_dim)
        trash_x = torch.randn(3, 2, spatial_dim)
        trash_y = decoder(encoder(trash_x))
        self.assertEqual(trash_x.shape, trash_y.shape)

    @given(
        spatial_exponent=st.integers(min_value=6, max_value=14),
        filters_in=st.integers(min_value=1, max_value=128),
        filters_out=st.integers(min_value=1, max_value=128),
        stride=st.integers(min_value=1, max_value=8),
        expand_ratio=st.integers(min_value=1, max_value=10),
        squeeze_ratio=st.integers(min_value=1, max_value=10),
        kernel_size=st.integers(min_value=1, max_value=31),
    )
    def test_io_gblock(self, spatial_exponent, filters_in, filters_out, stride, expand_ratio, squeeze_ratio, kernel_size):
        spatial_size = 2**spatial_exponent
        if kernel_size % 2 == 0:
            # kernel_size must be odd
            kernel_size += 1
        if kernel_size < stride:
            # blockgen would not allow this
            return
        while squeeze_ratio > filters_in * expand_ratio:
            # blockgen would not allow this
            squeeze_ratio -= 1
        squeeze_ratio = max(filters_in * expand_ratio, squeeze_ratio)
        blk = GBlock(
            filters_in=filters_in, filters_out=filters_out,
            stride=stride, kernel_size=kernel_size,
            expand_ratio=expand_ratio, squeeze_ratio=squeeze_ratio
        )
        trash_x = torch.randn(2, filters_in, spatial_size)
        trash_y = blk(trash_x)
        spatial_out = int((spatial_size - kernel_size + 2 * blk.pad_size) / stride + 1)
        # check for output shape correct
        self.assertEqual(trash_y.shape, (2, filters_out, spatial_out))
        loss = torch.nn.MSELoss()(trash_y, torch.randn(trash_y.shape))
        # will raise RuntimeError here if there is an issue with backprop
        loss.backward()

    @given(
        exponent_in=st.integers(min_value=2, max_value=14),
        exponent_out=st.integers(min_value=2, max_value=14),
        steps=st.integers(min_value=1, max_value=5),
        quantize_in=st.booleans(), quantize_out=st.booleans(),
        use_dropout=st.booleans()
    )
    def test_io_fc(self, exponent_in, exponent_out, steps, quantize_in, quantize_out, use_dropout):
        '''block should work with a variety of configs'''
        size_in, size_out = exponent_in**2, exponent_out**2
        autoencoder = FullyConnected(
            size_in=size_in, size_out=size_out,
            steps=steps, quantize_in=quantize_in, quantize_out=quantize_out
        )
        trash_x = torch.randn(3, size_in)
        trash_y = autoencoder(trash_x)
        self.assertEqual(trash_y.shape[1], size_out)



'''ensure layers are working'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import numpy as np

import torch

from glaucus import (
    TimeDomain2FreqDomain,
    FreqDomain2TimeDomain,
    RMSNormalize,
    GaussianNoise,
    RFLoss,
)


class TestDomainTransforms(unittest.TestCase):
    def test_roundtrip(self):
        '''
        Time -> Freq -> Time should yield identical results
        Freq -> Time -> Freq should yield identical results
        '''
        batch_size = np.random.randint(0, 64)
        spatial_size = 2**np.random.randint(8, 16)
        original = torch.rand(batch_size, 2, spatial_size)
        layer_f2t = FreqDomain2TimeDomain()
        layer_t2f = TimeDomain2FreqDomain()
        roundtrip_ftf = layer_f2t(layer_t2f(original))
        roundtrip_tft = layer_t2f(layer_f2t(original))
        self.assertTrue(torch.allclose(roundtrip_ftf, original, atol=1e-5))
        self.assertTrue(torch.allclose(roundtrip_tft, original, atol=1e-5))


class TestNormalization(unittest.TestCase):
    def test_rms_normalize(self):
        '''
        Tests the RMSNormalize layer to ensure the layer is normalizing inputs to RMS.
        '''
        batch_size = np.random.randint(1, 64)
        spatial_size = 2**np.random.randint(8, 16)
        # generate batches with different means and stdevs
        means = np.geomspace(1e-2, 1e3, 6) * (np.random.randint(0, 2, size=6) * 2 - 1)
        stdevs = np.geomspace(1e-2, 1e8, 4)
        layer = RMSNormalize(spatial_size=spatial_size)
        for mean in means:
            for stdev in stdevs:
                x_ncl = mean + torch.randn(batch_size, 2, spatial_size) * stdev
                y_ncl = layer(x_ncl)
                # convert from reals to complex for analysis
                y_nl = torch.view_as_complex(y_ncl.swapaxes(-1, -2).contiguous())
                # recall that to calculate true RMS you can either use this, norm, or vdot
                y_rms = torch.sqrt(torch.mean(torch.square(torch.abs(y_nl)), axis=-1))
                self.assertTrue(torch.allclose(y_rms, torch.ones(batch_size), rtol=1e-2))


class TestGaussianNoise(unittest.TestCase):
    def test_skip_on_eval(self):
        '''
        When self.training == True (before eval) noise will be added with this layer.
        Otherwise it will just return the same input.
        '''
        noise_layer = GaussianNoise(spatial_size=64)
        alpha = torch.randn(1, 2, 64)
        omega, _ = noise_layer(alpha)
        self.assertFalse(torch.equal(alpha, omega))
        noise_layer.eval()
        omega, _ = noise_layer(alpha)
        self.assertTrue(torch.equal(alpha, omega))

    def test_snr_ranges(self):
        '''lower the SNR, lower the relationship to original signal'''
        alpha = torch.randn(1, 2, 64)
        rfloss = RFLoss(weight_spec=0)
        for min_snr_db in np.arange(-10, 15, 5):
            high_noise_layer = GaussianNoise(
                spatial_size=64,
                min_snr_db=min_snr_db,
                max_snr_db=min_snr_db+1)
            low_noise_layer = GaussianNoise(
                spatial_size=64,
                min_snr_db=min_snr_db+5,
                max_snr_db=min_snr_db+6)
            omega_high, _ = high_noise_layer(alpha)
            omega_low, _ = low_noise_layer(alpha)
            self.assertLess(
                rfloss(alpha, omega_low)[0],
                rfloss(alpha, omega_high)[0]
            )



'''ensure RFLoss is working'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest
import torch

from glaucus import RFLoss


class TestRFLoss(unittest.TestCase):
    def _gen_x(self, batch_size=32, spatial_size=4096) -> None:
        self.alpha = torch.randn(batch_size, 2, spatial_size)
        self.omega = torch.randn(batch_size, 2, spatial_size)

    def test_weights(self):
        '''
        make sure returns correct number of values
        zero out weights for keys one by one
        should always return individual loss and the total loss
        '''
        weight_keys = ['weight_env', 'weight_fft', 'weight_align', 'weight_spec', 'weight_xcor']
        self._gen_x()
        for test_key in weight_keys:
            kwargs = {test_key: 1}
            for zero_key in weight_keys:
                if zero_key != test_key:
                    kwargs[zero_key] = 0

            criterion = RFLoss(**kwargs)
            _, metrics = criterion(self.alpha, self.omega)
            self.assertEqual(2, len(metrics))

    def test_naive_case(self):
        '''loss between identical signals should be near zero'''
        for spatial_size in 2**torch.arange(8, 14):
            self._gen_x(spatial_size=spatial_size)
            criterion = RFLoss(spatial_size=spatial_size)
            total_loss, _ = criterion(self.alpha, self.alpha)
            self.assertAlmostEqual(total_loss.numpy(), 0, places=5)

    def test_spec_loss(self):
        '''
        the spec_loss weight is scaled by spatial_size,
        so make sure that's working as intended
        '''
        for spatial_size in 2**torch.arange(8, 14):
            self._gen_x(spatial_size=spatial_size)
            criterion = RFLoss(spatial_size=spatial_size)
            # should be around 1 for uncorrelated inputs
            _, metrics = criterion(self.alpha, self.omega)
            self.assertAlmostEqual(metrics['spec_loss'].numpy(), 0.858, places=1)
            # should be 0 for equal inputs
            _, metrics = criterion(self.alpha, self.alpha)
            self.assertAlmostEqual(metrics['spec_loss'].numpy(), 0)
            # should be 0 for inversely correlated inputs due to absolute value
            _, metrics = criterion(self.alpha, -self.alpha)
            self.assertAlmostEqual(metrics['spec_loss'].numpy(), 0, places=2)

    def test_fft_loss(self):
        for spatial_size in 2**torch.arange(8, 14):
            criterion = RFLoss(spatial_size=spatial_size, weight_fft=1)
            # create a pair of slices with different AM tones
            alpha = torch.vstack((torch.sin(torch.arange(spatial_size)), torch.zeros(spatial_size))).unsqueeze(0)
            omega = torch.vstack((torch.sin(torch.arange(spatial_size) * .1), torch.zeros(spatial_size))).unsqueeze(0)
            # fft_loss should be very high
            _, metrics = criterion(alpha, omega)
            self.assertGreater(metrics['fft_loss'], 0.95)



#!/usr/bin/env python3
'''
  ________.__
 /  _____/|  | _____   __ __   ____  __ __  ______
/   \  ___|  | \__  \ |  |  \_/ ___\|  |  \/  ___/
\    \_\  \  |__/ __ \|  |  /\  \___|  |  /\___ \
 \______  /____(____  /____/  \___  >____//____  >
        \/          \/            \/           \/
'''
import os
import re
from setuptools import setup

with open(os.path.join('glaucus', '__init__.py'), encoding='utf-8') as derp:
    version = re.search(r'__version__\s*=\s*[\'"]([^\'"]*)[\'"]', derp.read()).group(1)

with open('README.md') as derp:
    long_description = derp.read()

setup(
    name='glaucus',
    version=version,
    author='Kyle Logue',
    author_email='kyle.logue@aero.org',
    url='https://github.com/the-aerospace-corporation/glaucus',
    license='GNU Lesser General Public License v3 or later (LGPLv3+)',
    test_suite='tests',
    packages=['glaucus'],
    description='Complex-valued encoder, decoder, and loss for RF DSP in PyTorch.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=[
        'torch',        # basic ML framework
        'lightning',    # extensions for PyTorch
        'madgrad',      # our favorite optimizer
        'hypothesis',   # best unit testing
    ],
)



# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import logging

import torch
import lightning as pl
from madgrad import MADGRAD

from .gblocks import GlaucusNet, ENCODER_BLOCKS, DECODER_BLOCKS
from .fcblocks import FullyConnected
from .layers import RMSNormalize, GaussianNoise, TimeDomain2FreqDomain, FreqDomain2TimeDomain
from .rfloss import RFLoss

log = logging.getLogger(__name__)


class GlaucusAE(pl.LightningModule):
    '''RF Autoencoder constructed with network of GBlocks.'''
    def __init__(self, encoder_blocks=ENCODER_BLOCKS, decoder_blocks=DECODER_BLOCKS,
            domain:str='time', width_coef:float=1, depth_coef:float=1, spatial_size:int=4096,
            bottleneck_in:int=512, bottleneck_latent:int=512, bottleneck_out:int=512,
            bottleneck_steps:int=1, bottleneck_quantize:bool=False, data_format:str='ncl',
            drop_connect_rate:float=0.2, optimizer:str='madgrad', lr:float=1e-3,
            ) -> None:
        '''
        encoder_blocks : list of namedtuple, optional
            Parameters to define sequential neural net architecture. Note that encoder output shape should be compatible
            with bottleneck parameters.
        decoder_blocks : list of namedtuple, optional
            See encoder blocks.
        domain : str, default time
            Force the encoder/decoder to operate in either time or freq domain.
        width_coef : float, default 1
            Coefficient to scale the width of the encoder/decoder blocks providing more channels for a given architecture.
        depth_coef : float, default 1
            Coefficient to scale the depth of the encoder/decoder blocks increasing complexity (and hopefully capacity).
        spatial_size : int, default 4096
            This network consumes RF data in NCL format. In order to properly do normalization properly and keep losses
            consistent we need to know the input length up front.
        bottleneck_in : int, default 512
            After feature extraction we enter a bottleneck layer. The input size of this layer should be equal to the
            output size of the encoder defined in `encoder_blocks`.
        bottleneck_latent : int, default 512
            The smallest point of the bottleneck that defines the size of our latent space.
        bottleneck_out : int, default 512
            The size of the bottleneck output should be equal to the input size of the decoder defined within
            `decoder_blocks`.
        bottleneck_step : int, default 1
            Within the bottleneck this determines how many steps down/up to/from the latent space.
        bottleneck_quantize : bool, default False
            This option adds a quant/dequant step in the latent space for maximum compression.
        drop_connect_rate : float, optional
            Drop connect is a generalization of dropout that allows better regularization for deeper networks.
            As implemented the rate will scale from 0 for the first layer up to the provided drop_connect_rate for the
            final layer. `drop_connect_rate` is the inverse of `survival_rate`.
            Very large models can generalize training with values up to 0.5, but in Dec 2021
            a comparison between using max of 0.2 and a scaled max value (0.2 to 0.5) depending on model size
            determined the larger rates only extended training time and didn't generally improve. Stick with 0.2.
        optimizer : string, default madgrad
            Currently support either `madgrad` or `adam` optimizers.
        lr : float, default 1e-3
            Learning Rate. Experiments from Dec 2021 to Mar 2022 yielded good values in range (1e-3, 1e-2).
        data_format : str, default 'ncl'
            Network normally consumes and produces complex-valued data represented as real-valued (NCL)
            but if data is complex-valued (NL) will add a transform layer during encode/decode.
        '''
        super().__init__()

        self.save_hyperparameters()
        self.lr = lr
        assert domain in ['time', 'freq']
        assert data_format in ['ncl', 'nl']
        self.domain = domain
        self.data_format = data_format

        self._rms_norm = RMSNormalize(spatial_size=spatial_size)
        self._noise_layer = GaussianNoise(spatial_size=spatial_size)
        if self.domain == 'freq':
            self._time2freq = TimeDomain2FreqDomain()
            self._freq2time = FreqDomain2TimeDomain()
        self.loss_function = RFLoss(spatial_size=spatial_size, data_format=data_format)
        self.encoder = GlaucusNet(
            encoder_blocks, mode='encoder', width_coef=width_coef, depth_coef=depth_coef, drop_connect_rate=drop_connect_rate)
        self.fc_encoder = FullyConnected(
            size_in=bottleneck_in, size_out=bottleneck_latent, quantize_out=bottleneck_quantize, steps=bottleneck_steps)
        self.fc_decoder = FullyConnected(
            size_in=bottleneck_latent, size_out=bottleneck_out, quantize_in=bottleneck_quantize, steps=bottleneck_steps)
        self.decoder = GlaucusNet(
            decoder_blocks, mode='decoder',
            width_coef=width_coef, depth_coef=depth_coef, drop_connect_rate=drop_connect_rate)
        optimizer_map = {'adam': torch.optim.Adam, 'madgrad': MADGRAD}
        self.optimizer = optimizer_map[optimizer]

        log.info('GlaucusAE(%s)', self.domain)

    def forward(self, x):
        x_enc = self.encode(x)
        x_hat = self.decode(x_enc)
        return x_hat, x_enc

    def encode(self, x):
        '''normalize, add noise if training, and reduce to latent domain'''
        if self.data_format == 'nl':
            x = torch.view_as_real(x).swapaxes(-1,-2)
        x = self._rms_norm(x)
        x, _ = self._noise_layer(x)
        if self.domain == 'freq':
            # convert to frequency domain
            x = self._time2freq(x)
        feats = self.encoder(x)
        x_enc = self.fc_encoder(feats)
        return x_enc

    def decode(self, x_enc):
        '''return from latent domain to complex RF'''
        feats = self.fc_decoder(x_enc)
        x_hat = self.decoder(feats)
        if self.domain == 'freq':
            # convert back to time domain
            x_hat = self._freq2time(x_hat)
        if self.data_format == 'nl':
            x_hat = torch.view_as_complex(x_hat.swapaxes(-1,-2).contiguous())
        return x_hat

    def step(self, batch, batch_idx):
        x, metadata = batch
        x_hat, _ = self.forward(x)
        loss, metrics = self.loss_function(x_hat, x)
        return loss, metrics

    def training_step(self, batch, batch_idx):
        loss, _ = self.step(batch, batch_idx)
        self.log('train_loss', loss, on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, metrics = self.step(batch, batch_idx)
        self.log_dict({f"val_{k}": v for k, v in metrics.items()}, sync_dist=True)
        return loss

    def test_step(self, batch, batch_idx):
        loss, metrics = self.step(batch, batch_idx)
        self.log_dict({f"test_{k}": v for k, v in metrics.items()}, sync_dist=True)
        return metrics

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters(), lr=self.lr)
        return optimizer


class FullyConnectedAE(pl.LightningModule):
    '''RF Autoencoder constructed with fully connected layers.'''
    def __init__(self,
        spatial_size:int=4096, latent_dim:int=512, lr:float=1e-3, steps:int=3, bottleneck_quantize:bool=False,
        domain:str='time', data_format:str='ncl', optimizer:str='madgrad') -> None:
        super().__init__()
        self.save_hyperparameters()
        self.lr = lr
        assert domain in ['time', 'freq']
        assert data_format in ['ncl', 'nl']
        self.domain = domain
        self.data_format = data_format

        self.latent_dim = latent_dim
        self.io_dim = spatial_size * 2
        self.steps = steps

        self._rms_norm = RMSNormalize(spatial_size=spatial_size)
        self._noise_layer = GaussianNoise(spatial_size=spatial_size)
        if self.domain == 'freq':
            self._time2freq = TimeDomain2FreqDomain()
            self._freq2time = FreqDomain2TimeDomain()
        self.loss_function = RFLoss(spatial_size=spatial_size, data_format=data_format)

        optimizer_map = {'adam': torch.optim.Adam, 'madgrad': MADGRAD}
        self.optimizer = optimizer_map[optimizer]

        self.encoder = FullyConnected(
            size_in=self.io_dim, size_out=self.latent_dim, quantize_out=bottleneck_quantize, steps=steps, use_dropout=True)
        self.decoder = FullyConnected(
            size_in=self.latent_dim, size_out=self.io_dim, quantize_in=bottleneck_quantize, steps=steps, use_dropout=True)
        # replace final activation with Tanh
        self.decoder._fc[-1] = torch.nn.Tanh()

        log.info('FullyConnectedAE(%s)', self.domain)

    def forward(self, x):
        x_enc = self.encode(x)
        x_hat = self.decode(x_enc)
        return x_hat, x_enc

    def encode(self, x):
        '''normalize, add noise if training, and reduce to latent domain'''
        if self.data_format == 'nl':
            x = torch.view_as_real(x).swapaxes(-1, -2)
        x = self._rms_norm(x)
        x, _ = self._noise_layer(x)
        if self.domain == 'freq':
            # convert to frequency domain
            x = self._time2freq(x)
        # NCL to NL
        x = x.flatten(start_dim=1)
        x_enc = self.encoder(x)
        return x_enc

    def decode(self, x_enc):
        '''return from latent domain to complex RF'''
        x_hat = self.decoder(x_enc)
        # NL to NCL
        x_hat = x_hat.view(x_hat.size(0), 2, -1)
        if self.domain == 'freq':
            # convert back to time domain
            x_hat = self._freq2time(x_hat)
        if self.data_format == 'nl':
            x_hat = torch.view_as_complex(x_hat.swapaxes(-1, -2).contiguous())
        return x_hat

    def step(self, batch, batch_idx):
        x, metadata = batch
        x_hat, _ = self.forward(x)
        loss, metrics = self.loss_function(x_hat, x)
        return loss, metrics

    def training_step(self, batch, batch_idx):
        loss, _ = self.step(batch, batch_idx)
        # self.log_dict({f"val_{k}": v for k, v in metrics.items()})
        self.log('train_loss', loss, on_step=True, on_epoch=False)
        return loss

    def validation_step(self, batch, batch_idx):
        loss, metrics = self.step(batch, batch_idx)
        self.log_dict({f"val_{k}": v for k, v in metrics.items()})
        return loss

    def configure_optimizers(self):
        optimizer = self.optimizer(self.parameters(), lr=self.lr)
        return optimizer



'''version should be parsable'''
# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

import unittest

from glaucus import __version__


class TestVersion(unittest.TestCase):
    def test_version(self):
        '''
        Ensure the version string is valid.
        '''
        major, minor, micro = tuple(int(x) for x in __version__.split('.'))
        print(f'mod_classify v{major}.{minor}.{micro}')



# Copyright 2023 The Aerospace Corporation
# This file is a part of Glaucus
# SPDX-License-Identifier: LGPL-3.0-or-later

__version__ = '1.1.0'

from .rfloss import *
from .layers import *
from .gblocks import *
from .fcblocks import *
from .autoencoders import *
